@@ -1,20 +1,30 @@
+import logging
 from contextlib import contextmanager
-from typing import Any, Iterable, Iterator, List, Optional
+from typing import Any, Generic, Iterable, Iterator, Optional, TypeVar
 
 from pymongo import MongoClient
 from pymongo.client_session import ClientSession
 from pymongo.database import Database
 
-from apibara.model import EventFilter
+import apibara.cursor as cursor_utils
+from apibara.indexer.configuration import IndexerConfiguration
+from apibara.protocol.proto.stream_pb2 import Cursor
 
 Document = dict[str, Any]
-Filter = dict[str, Any]
+DocumentFilter = dict[str, Any]
 Update = dict[str, Any]
 Projection = dict[str, any]
 
+Filter = TypeVar("Filter")
 
-class IndexerStorage:
-    """Create instances of Storage."""
+
+logger = logging.getLogger(__name__)
+
+
+class IndexerStorage(Generic[Filter]):
+    """
+    Manage indexers storage.
+    """
 
     def __init__(self, url: Optional[str], indexer_id: str) -> None:
         if url is None:
@@ -27,134 +37,124 @@ class IndexerStorage:
         self.db = self._mongo[self.db_name]
 
     @contextmanager
-    def create_storage_for_block(self, block_number: int) -> Iterator["Storage"]:
+    def create_storage_for_block(self, cursor: Cursor) -> Iterator["Storage"]:
         with self._mongo.start_session() as session:
-            yield Storage(self.db, session, block_number)
+            yield Storage(self.db, session=session, cursor=cursor)
 
     @contextmanager
-    def create_storage_for_data(self, block_number: int) -> Iterator["Storage"]:
+    def create_storage_for_data(self, cursor: Cursor) -> Iterator["Storage"]:
         with self._mongo.start_session() as session:
-            yield Storage(self.db, session, block_number)
-            self._update_indexed_to(block_number, session)
+            yield Storage(self.db, session=session, cursor=cursor)
+            self._update_cursor(cursor, session)
 
     @contextmanager
-    def create_storage_for_invalidate(self, block_number: int) -> Iterator["Storage"]:
+    def create_storage_for_invalidate(self, cursor: Cursor) -> Iterator["Storage"]:
         with self._mongo.start_session() as session:
-            yield Storage(self.db, session, block_number)
-            if block_number > 0:
-                self._update_indexed_to(block_number - 1, session)
+            yield Storage(self.db, session=session, cursor=cursor)
+            self._update_cursor(cursor, session)
 
     @contextmanager
-    def create_storage_for_pending(self, block_number: int) -> Iterator["Storage"]:
+    def create_storage_for_pending(self, cursor: Cursor) -> Iterator["Storage"]:
         with self._mongo.start_session() as session:
-            yield Storage(self.db, session, block_number)
+            yield Storage(self.db, session=session, cursor=cursor)
 
-    def initialize(self, starting_sequence: int, filters: List[EventFilter]):
+    def _initialize_configuration(self, configuration: IndexerConfiguration[Filter]):
         existing = self.db["_apibara"].find_one({"indexer_id": self._indexer_id})
         if existing is not None:
             return
-        # the indexer hasn't really started indexing anything yet, but if it's
-        # stopped right before receiving any message it will restart from the
-        # next block, skipping the starting block.
-        indexed_to = starting_sequence - 1
+
+        logger.debug("writing initial configuration to storage")
+
+        cursor = None
+        if configuration.starting_cursor is not None:
+            cursor = cursor_utils.to_json(configuration.starting_cursor)
+
+        if configuration.filter is None:
+            raise RuntimeError("configuration filter must be defined")
+
+        filter = configuration.filter.encode()
 
         self.db["_apibara"].insert_one(
             {
                 "indexer_id": self._indexer_id,
-                "indexed_to": indexed_to,
-                "filters": [f.to_json() for f in filters],
+                "cursor": cursor,
+                "filter": filter,
             }
         )
 
-    def starting_sequence(self):
-        state = self.db["_apibara"].find_one({"indexer_id": self._indexer_id})
-        if state is None or state.get("indexed_to") is None:
-            return None
-        return state["indexed_to"] + 1
+    def update_with_stored_configuration(self, initial: IndexerConfiguration[Filter]):
+        logger.debug("update config with stored version")
 
-    def event_filters(self):
-        state = self.db["_apibara"].find_one({"indexer_id": self._indexer_id})
-        if state is None:
-            raise RuntimeError("indexer state not found")
-        return [EventFilter.from_json(j) for j in state["filters"]]
+        stored = self.db["_apibara"].find_one({"indexer_id": self._indexer_id})
+        if stored is None:
+            self._initialize_configuration(initial)
+            return False
+
+        encoded_filter = stored.get("filter")
+        if encoded_filter is None:
+            raise RuntimeError("indexer filter is missing")
+        initial.filter.parse(encoded_filter)
+
+        cursor = stored.get("cursor")
+        if cursor is not None:
+            initial.starting_cursor = cursor_utils.from_json(cursor)
+
+        return True
+
+    def _update_filter(self, filter: Filter, session: ClientSession):
+        """Set the indexer filter, overriding the previous filter."""
+        logger.debug("update stored filter")
 
-    def _set_event_filters(self, filters: List[EventFilter], session: ClientSession):
-        """Set the indexer event filters, overriding the previous filters."""
-        filters_docs = [f.to_json() for f in filters]
         self.db["_apibara"].update_one(
             {"indexer_id": self._indexer_id},
-            {"$set": {"filters": filters_docs}},
+            {"$set": {"filter": filter.encode()}},
             session=session,
         )
 
-    def invalidate(self, block_number: int):
-        """Invalidates all data generate in `block_number` and later."""
+    def invalidate(self, cursor: Cursor):
+        """Invalidates all data generate after `cursor`."""
+        logger.debug("invalidate data after %A", cursor)
         with self._mongo.start_session() as session:
             for collection in self.db.list_collections(session=session):
                 name = collection["name"]
                 if name.startswith("_"):
                     continue
-                # remove items inserted in or after block_number
+                # remove items inserted after block_number
                 self.db[name].delete_many(
-                    {"_chain.valid_from": {"$gte": block_number}}, session=session
+                    {"_chain.valid_from": {"$gt": cursor.order_key}}, session=session
                 )
-                # rollback items updated in or after block_number
+                # rollback items updated after block_number
                 self.db[name].update_many(
-                    {"_chain.valid_to": {"$gte": block_number}},
+                    {"_chain.valid_to": {"$gt": cursor.order_key}},
                     {"$set": {"_chain.valid_to": None}},
                     session=session,
                 )
 
     def drop_database(self):
+        logger.debug("dropping database %s", self.db_name)
         self._mongo.drop_database(self.db_name)
 
-    def _update_indexed_to(self, block_number: int, session: ClientSession):
+    def _update_cursor(self, cursor: Cursor, session: Optional[ClientSession] = None):
         self.db["_apibara"].update_one(
             {"indexer_id": self._indexer_id},
-            {"$set": {"indexed_to": block_number}},
+            {"$set": {"cursor": cursor_utils.to_json(cursor)}},
             upsert=True,
             session=session,
         )
 
 
-class Storage:
-    """Chain-aware document storage."""
+class ReadOnlyStorage:
+    """Chain-aware document storage, read methods."""
 
-    def __init__(self, db: Database, session: ClientSession, block_number: int) -> None:
+    def __init__(
+        self, db: Database, *, session: Optional[ClientSession] = None
+    ) -> None:
         self._db = db
-        self._block_number = block_number
         self._session = session
 
-    async def insert_one(self, collection: str, doc: Document):
-        """Insert `doc` into `collection`."""
-        self._add_chain_information(doc)
-        self._db[collection].insert_one(doc, session=self._session)
-
-    async def insert_many(self, collection: str, docs: Iterable[Document]):
-        """Insert multiple `docs` into `collection`."""
-        for doc in docs:
-            self._add_chain_information(doc)
-        self._db[collection].insert_many(docs, session=self._session)
-
-    async def delete_one(self, collection: str, filter: Filter):
-        """ "Delete the first document in `collection` matching `filter`."""
-        self._add_current_block_to_filter(filter)
-        self._db[collection].update_one(
-            filter,
-            {"$set": {"_chain.valid_to": self._block_number}},
-            session=self._session,
-        )
-
-    async def delete_many(self, collection: str, filter: Filter):
-        """Delete all documents in `collection` matching `filter`."""
-        self._add_current_block_to_filter(filter)
-        self._db[collection].update_many(
-            filter,
-            {"$set": {"_chain.valid_to": self._block_number}},
-            session=self._session,
-        )
-
-    async def find_one(self, collection: str, filter: Filter) -> Optional[Document]:
+    async def find_one(
+        self, collection: str, filter: DocumentFilter
+    ) -> Optional[Document]:
         """Find the first document in `collection` matching `filter`."""
         self._add_current_block_to_filter(filter)
         return self._db[collection].find_one(filter, session=self._session)
@@ -162,7 +162,7 @@ class Storage:
     async def find(
         self,
         collection: str,
-        filter: Filter,
+        filter: DocumentFilter,
         sort: Optional[dict[str, int]] = None,
         projection: Optional[Projection] = None,
         skip: int = 0,
@@ -188,22 +188,63 @@ class Storage:
                 cursor = cursor.sort(field, order)
         return cursor
 
+    def _add_current_block_to_filter(self, filter: DocumentFilter):
+        filter["_chain.valid_to"] = None
+
+
+class Storage(ReadOnlyStorage):
+    """Chain-aware document storage."""
+
+    def __init__(
+        self, db: Database, cursor: Cursor, *, session: Optional[ClientSession] = None
+    ) -> None:
+        super().__init__(db, session=session)
+        self._cursor = cursor
+
+    async def insert_one(self, collection: str, doc: Document):
+        """Insert `doc` into `collection`."""
+        self._add_chain_information(doc)
+        self._db[collection].insert_one(doc, session=self._session)
+
+    async def insert_many(self, collection: str, docs: Iterable[Document]):
+        """Insert multiple `docs` into `collection`."""
+        for doc in docs:
+            self._add_chain_information(doc)
+        self._db[collection].insert_many(docs, session=self._session)
+
+    async def delete_one(self, collection: str, filter: DocumentFilter):
+        """ "Delete the first document in `collection` matching `filter`."""
+        self._add_current_block_to_filter(filter)
+        self._db[collection].update_one(
+            filter,
+            {"$set": {"_chain.valid_to": self._cursor.order_key}},
+            session=self._session,
+        )
+
+    async def delete_many(self, collection: str, filter: DocumentFilter):
+        """Delete all documents in `collection` matching `filter`."""
+        self._add_current_block_to_filter(filter)
+        self._db[collection].update_many(
+            filter,
+            {"$set": {"_chain.valid_to": self._cursor.order_key}},
+            session=self._session,
+        )
+
     async def find_one_and_replace(
         self,
         collection: str,
-        filter: Filter,
+        filter: DocumentFilter,
         replacement: Document,
         upsert: bool = False,
     ):
         """Replace the first document in `collection` matching `filter` with `replacement`.
-
         If `upsert = True`, insert `replacement` even if no document matched the `filter`.
         """
         # Step 1. Update the old document (if any) by clamping its validity range
         self._add_current_block_to_filter(filter)
         existing = self._db[collection].find_one_and_update(
             filter,
-            {"$set": {"_chain.valid_to": self._block_number}},
+            {"$set": {"_chain.valid_to": self._cursor.order_key}},
             session=self._session,
         )
 
@@ -215,14 +256,14 @@ class Storage:
         return existing
 
     async def find_one_and_update(
-        self, collection: str, filter: Filter, update: Update
+        self, collection: str, filter: DocumentFilter, update: Update
     ):
         """Update the first document in `collection` matching `filter` with `update`."""
         # Step 1. Update the old document (if any) by clamping its validity range
         self._add_current_block_to_filter(filter)
         existing = self._db[collection].find_one_and_update(
             filter,
-            {"$set": {"_chain.valid_to": self._block_number}},
+            {"$set": {"_chain.valid_to": self._cursor.order_key}},
             session=self._session,
         )
 
@@ -236,7 +277,4 @@ class Storage:
         return existing
 
     def _add_chain_information(self, doc: Document):
-        doc["_chain"] = {"valid_from": self._block_number, "valid_to": None}
-
-    def _add_current_block_to_filter(self, filter: Filter):
-        filter["_chain.valid_to"] = None
+        doc["_chain"] = {"valid_from": self._cursor.order_key, "valid_to": None}
